//Codec for this file is Windows-1251
//C++17

#ifndef _WIN32_WINNT
#define _WIN32_WINNT 0x0500
#endif // _WIN32_WINNT

#ifndef UNLEN
#define UNLEN 256
#endif // UNLEN

#include "FuncWrapper.h"
#include "Helper.h"

#include <sstream>
#include <winnls.h>
#include <winreg.h>
#include <winver.h>
#include <tlhelp32.h>

#include <winspool.h>
#include <psapi.h>
#include <setupapi.h>
//  libwinspool.a libpsapi.a  libsetupapi.a

namespace{

    constexpr char
        ERROR_CANT_ALLOC_MEM[]{".    ."};
    constexpr ::DWORD MAX_BUF{1024}; // 

}

std::string FuncWrapper::err_description(std::string_view fun_name) {
    const ::DWORD err_code = ::GetLastError();
    constexpr char CANT_ALLOC[]{"<     >"};
    constexpr char SPACE_ERR_RET_SPACE[]{"    "};
    constexpr char DOT_SPACE[]{". "};
    HeapBuffer<char> sysMsgBuf{::MAX_BUF};
    std::string sysMsg;
    if( sysMsgBuf.isValid() ) {
        ::FormatMessageA(
            FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL,
                err_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                    sysMsgBuf.get(), sysMsgBuf.ItemsCount(), NULL);
        sysMsg = sysMsgBuf.get();
        sysMsg.resize(sysMsg.length()-2); //  
    } else
        sysMsg = CANT_ALLOC;
    std::stringstream ss;
    ss << std::string{fun_name} << SPACE_ERR_RET_SPACE << err_code
        << DOT_SPACE << sysMsg;
    return ss.str();
}

bool PlatformInfo::exec() {
    ::OSVERSIONINFOA vi;
    vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOA); // 
    if( !::GetVersionExA(&vi) ) {
        err_str = err_description("GetVersionExA");
        return false;
    }
    const ::DWORD pltID = vi.dwPlatformId;
    std::stringstream ss;
    ss << "ID : " << pltID;
    switch( pltID ) {
    case VER_PLATFORM_WIN32s:
        ss << " (Win32s  Windows 3.1)";
        break;
    case VER_PLATFORM_WIN32_WINDOWS:
        ss << " (Win32  Windows 95/98/Me)";
        break;
    case VER_PLATFORM_WIN32_NT:
        ss << " (Win32NT  Windows NT/2000/XP...)";
        break;
    default:
        ss << " (.)";
    }
    ss << "\n  : " << vi.dwMajorVersion << "."
       << vi.dwMinorVersion << "\n";
    ::DWORD bn = 0;
    if( vi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ) //Win9x/Me
        bn = LOWORD(vi.dwBuildNumber);
    else
        bn = vi.dwBuildNumber; //
    ss << " : " << bn << "\n";
    ss << "SP ( WinNT,  ): " << vi.szCSDVersion;
    res_str = ss.str();
    return true;
}

bool Language::exec() {
    HeapBuffer<char> buf{MAX_PATH+1};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !::VerLanguageNameA(::GetSystemDefaultLangID(), buf.get(),
        buf.ItemsCount()) ) {
            err_str = err_description("VerLanguageNameA");
            return false;
    }
    res_str = ": " + std::string(buf.get());
    return true;
}

bool UserName::exec() {
    HeapBuffer<char> buf{UNLEN+1};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    ::DWORD ln = buf.ItemsCount();
    if ( !::GetUserNameA(buf.get(), &ln) ) {
        err_str = err_description("GetUserNameA");
        return false;
    }
    res_str = " : " + std::string(buf.get());
    return true;
}

bool CompName::exec() {
    HeapBuffer<char> buf{MAX_COMPUTERNAME_LENGTH+1};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    ::DWORD ln = buf.ItemsCount();
    if( !::GetComputerNameA(buf.get(), &ln) ) {
        err_str = err_description("GetComputerNameA");
        return false;
    }
    res_str = " : " + std::string(buf.get());
    return true;
}

bool WinDir::exec() {
    HeapBuffer<char> buf{MAX_PATH};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !::GetWindowsDirectoryA(buf.get(), buf.ItemsCount()) ) {
        err_str = err_description("GetWindowsDirectoryA");
        return false;
    }
    res_str = " Windows: " + std::string(buf.get());
    return true;
}

bool SysDir::exec() {
    HeapBuffer<char> buf{1024}; // 
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !GetSystemDirectoryA(buf.get(), buf.ItemsCount()) ) {
        err_str = err_description("GetSystemDirectoryA");
        return false;
    }
    res_str = " : " + std::string(buf.get());
    return true;
}

bool CurrDir::exec() {
    const ::DWORD needed = ::GetCurrentDirectoryA(0, NULL);
    HeapBuffer<char> buf{needed};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !::GetCurrentDirectoryA(buf.ItemsCount(), buf.get()) ) {
        err_str = err_description("GetCurrentDirectoryA");
        return false;
    }
    res_str = " : " + std::string(buf.get());
    return true;
}

bool GlobMemStEx::exec() {
    ::MEMORYSTATUSEX msx;
    msx.dwLength = sizeof msx; //   
    if( !::GlobalMemoryStatusEx(&msx) ) {
        err_str = err_description("GlobalMemoryStatusEx");
        return false;
    }
    std::stringstream ss;
    ss << "   GlobalMemoryStatusEx:\n"
       << " : " << msx.dwMemoryLoad << "%\n"
       << "  ():\n" // >> 20 --   
       << (::DWORD)(msx.ullTotalPhys >> 20) << " (),\n"
       << (::DWORD)(msx.ullTotalPageFile >> 20) << " (.),\n"
       << (::DWORD)(msx.ullTotalVirtual >> 20) << " (.)\n"
       << "  ():\n"
       << (::DWORD)(msx.ullAvailPhys >> 20) << " (),\n"
       << (::DWORD)(msx.ullAvailPageFile >> 20) << " (.),\n"
       << (::DWORD)(msx.ullAvailVirtual >> 20) << " (.)";
    //. MSDN    
    //   - ?
    ss << "\n:  .\
        \n  .";
    res_str = ss.str();
    return true;
}

bool PerfInf::exec() {
    ::PERFORMANCE_INFORMATION pInf;
    pInf.cb = sizeof(::PERFORMANCE_INFORMATION);//  
    if( !::GetPerformanceInfo(&pInf, pInf.cb) ) {
        err_str = err_description("GetPerformanceInfo");
        return false;
    }
    std::stringstream ss;
    ss << "   GetPerformanceInfo:\n"
       << " ():\n"
       << (pInf.PageSize*pInf.PhysicalTotal >> 20) << " (),\n"
       << (pInf.PageSize*pInf.PhysicalAvailable >> 20) << " ()\n"
       << " :\n"
       << " - " << pInf.ProcessCount << ",\n"
       << " - " << pInf.ThreadCount << ",\n"
       << ". - " << pInf.HandleCount;
    res_str = ss.str();
    return true;
}

bool Intel_x86_ProcArch::exec() { //  Intel x86 
    ::SYSTEM_INFO si;
    constexpr unsigned short ER_AR{12345}; //  
    si.wProcessorArchitecture = ER_AR;
    ::GetSystemInfo(&si);
    if( ER_AR == si.wProcessorArchitecture ) {
        err_str = " GetSystemInfo";
        return false;
    }
    if( PROCESSOR_ARCHITECTURE_INTEL != si.wProcessorArchitecture ) {
        err_str = "  Intel x86 .";
        return false;
    }
    int PrLev = si.wProcessorLevel;
    std::stringstream ss;
    ss << " GetSystemInfo:\n" << "(Intel)  : "
        << PrLev;
    switch (PrLev) {
    case 3:
        ss << " (80386)";
        break;
    case 4:
        ss << " (80486)";
        break;
    case 5:
        ss << " (Pentium)";
        break;
    case 6:
        ss << " (Pentium Pro)";
        break;
    default:
        ss << " (.)";
    }
    res_str = ss.str();
    return true;
}

bool VideoCard::exec() {
    ::DISPLAY_DEVICEA dd;
    dd.cb = sizeof(::DISPLAY_DEVICEA);
    if( !::EnumDisplayDevicesA(NULL, 0, &dd, 0) ) {
        err_str = " EnumDisplayDevicesA";
        return false;
    }
    res_str = ": " + std::string(dd.DeviceString);
    return true;
}

bool DefaultPrinter::exec() {
    ::DWORD needed;
    (void)::GetDefaultPrinterA(NULL, &needed);
    HeapBuffer<char> buf{needed};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !::GetDefaultPrinterA(buf.get(), &needed) ) {
        err_str = " GetDefaultPrinterA";
        return false;
    }
    res_str = " -: " + std::string(buf.get());
    return true;
}

bool PrintersInfo::exec() {
    ::DWORD Returned; // 
    ::DWORD Needed;   // /   
    constexpr ::DWORD flags = PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS;
    //
    ::BYTE tmp;
    //    ,   
    (void)::EnumPrintersA(flags, NULL, 2, &tmp, sizeof tmp, &Needed,
        &Returned);
    //   
    HeapBuffer<::BYTE> buf{Needed};
    if( !buf.isValid() ) {
        err_str = ::ERROR_CANT_ALLOC_MEM;
        return false;
    }
    if( !::EnumPrintersA(flags, NULL, 2, buf.get(), buf.BytesCount(), &Needed,
        &Returned) ) {
            err_str = " EnumPrintersA";
            return false;
    }
    if( !Returned ) {
        res_str = ": \n  .";
        return true;
    }
    std::stringstream ss;
    ss << ":\n";
    for(::DWORD i = 0; i < Returned; ++i)
        ss << ((::PRINTER_INFO_2*)buf.get())[i].pPrinterName <<
            " ( . " <<
                ((::PRINTER_INFO_2*)buf.get())[i].pDevMode->dmDriverVersion
                    << ")\n";
    res_str = ss.str();
    res_str.resize(res_str.length()-1);
    return true;
}

bool DiscsTypes::exec() {
    char dn[]= ".:";
    res_str = ":\n";
    for(char i = 'A'; i <= 'Z'; ++i) {
        dn[0] = i;
        ::UINT typD = ::GetDriveTypeA(dn);
        if( DRIVE_NO_ROOT_DIR != typD ) {
            res_str += dn;
            switch( typD ) {
            case DRIVE_FIXED :
                res_str += " ()";
                break;
            case DRIVE_CDROM :
                res_str += " ( CD/DVD)";
                break;
            case DRIVE_REMOVABLE :
                res_str += " ()";
                break;
            case DRIVE_REMOTE :
                res_str += " ()";
                break;
            case DRIVE_RAMDISK :
                res_str += " (RAM)";
                break;
            case DRIVE_UNKNOWN :
                res_str += " (.)";
                break;
            default :
                res_str += " (?)";
            }
            res_str += "\n";
        }
    }
    res_str.resize(res_str.length()-1);
    return true;
}

bool Procs_by_TlHelp32::exec() {
    const ::HANDLE hSnap = ::CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
    if( INVALID_HANDLE_VALUE == hSnap ) {
        err_str = err_description("CreateToolhelp32Snapshot");
        return false;
    }
    std::stringstream ss;
    ss << "  (  TlHelp32):\n";
    int i{0};
    ::DWORD th_cnt{0};
    ::PROCESSENTRY32 pe;
    pe.dwSize = sizeof pe;
    if( ::Process32First(hSnap, &pe) )
        do {
                ss << "#" << ++i << ": " << pe.szExeFile << "\n";
                th_cnt += pe.cntThreads;
        } while( ::Process32Next(hSnap, &pe) );
    (void)::CloseHandle(hSnap);
    ss << " : " << i << "\n"
       << " ( ?): " << th_cnt;
    res_str = ss.str();
    return true;
}

bool Procs_by_Psapi::exec() {
    //     SeDebugPrivilege,
    //    
    bool bDebug{false};
    if( ::ImpersonateSelf(::SecurityImpersonation) ) {
        ::HANDLE hToken;
        if( ::OpenThreadToken(::GetCurrentThread(),
            TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, FALSE, &hToken) ) {
            ::TOKEN_PRIVILEGES tp;
            ::LUID luid;
            if( ::LookupPrivilegeValueA(NULL, "SeDebugPrivilege", &luid) ) {
                tp.PrivilegeCount = 1;
                tp.Privileges[0].Luid = luid;
                tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
                bDebug = ::AdjustTokenPrivileges(hToken, FALSE, &tp,
                    sizeof(TOKEN_PRIVILEGES), NULL, NULL);
            }
            ::CloseHandle(hToken); //  
        }
    }
    std::stringstream ss;
    ss << " (  Psapi, ";
    if( bDebug )
        ss << " ";
    else
        ss << "   ";
    ss << "):\n";

    constexpr ::DWORD PROCS_CNT{2000}; //2000   
    HeapBuffer<::DWORD> procsIDs{PROCS_CNT};
    if( !procsIDs.isValid() ) {
        err_str = ".     .";
        return false;
    }
    ::DWORD bytesReturned;
    if( !::EnumProcesses(procsIDs.get(), procsIDs.BytesCount(),
        &bytesReturned) ) {
            err_str = err_description("EnumProcesses");
            return false;
    }
    const ::DWORD num = bytesReturned / sizeof(::DWORD); // 
    ::HANDLE procHndl;
    HeapBuffer<char> modName{::MAX_BUF};
    if( !modName.isValid() ) {
        err_str = ".     .";
        return false;
    }
    for(::DWORD i = 0; i < num; ++i) {
        ss << "#" << i+1 << " ID: " << procsIDs.get()[i] << " ";
        procHndl = ::OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_VM_READ,
            FALSE, procsIDs.get()[i]);
        if( !procHndl ) {
            ss << "? (" << err_description("OpenProcess") << ")\n";
            continue;
        }
        if( ::GetModuleFileNameExA(procHndl, NULL, modName.get(),
            modName.ItemsCount()) )
                ss << modName.get();
        else
            ss << "??? (" << err_description("GetModuleFileNameExA") << ")";
        ss << "\n";
        ::CloseHandle(procHndl);
    }
    ss << ":  .\
        \n   .   .";
    res_str = ss.str();
    return true;
}

bool DevicesList::exec() {
    HeapBuffer<char> Buffer{::MAX_BUF};
    if( !Buffer.isValid() ) {
        err_str = ".   .";
        return false;
    }
    const ::HDEVINFO lst = SetupDiGetClassDevsA(NULL, NULL, NULL,
        DIGCF_ALLCLASSES | DIGCF_PRESENT);
    if( INVALID_HANDLE_VALUE == lst) {
        err_str = err_description("SetupDiGetClassDevsA");
        return false;
    }
    res_str = " :\n";
    ::SP_DEVINFO_DATA DeviceInfoData;
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);//
    ::DWORD k{0};
    while( true ) {
        if( ::SetupDiEnumDeviceInfo(lst, k, &DeviceInfoData) ) {
            if( ::SetupDiGetDeviceRegistryPropertyA(lst, &DeviceInfoData,
                    SPDRP_FRIENDLYNAME, NULL, (PBYTE)Buffer.get(),
                        Buffer.BytesCount(), NULL) )
                            res_str += Buffer.get() + std::string("\n");
            else if( ::SetupDiGetDeviceRegistryPropertyA(lst, &DeviceInfoData,
                        SPDRP_DEVICEDESC, NULL, (PBYTE)Buffer.get(),
                            Buffer.BytesCount(), NULL) )
                                res_str += Buffer.get() + std::string("\n");
            else //    
                res_str += "*** ??? (   ) ***\n";
        } else if( ERROR_NO_MORE_ITEMS == ::GetLastError() )
            break;
        else //    
            res_str += "*** ??? (" + err_description("SetupDiEnumDeviceInfo")
                       + ") ***\n";
        k++;
    }
    ::SetupDiDestroyDeviceInfoList(lst); //  
    res_str.resize(res_str.length()-1);
    return true;
}
