//Codec for this file is Windows-1251
//C++17

//:  2020


#ifndef HELPER_H
#define HELPER_H

#include <type_traits>
#include <string_view>
#include <winbase.h>
#include <winuser.h>

void ApiWriteAsOem(const char* c_str);

template<typename POD_TYPE>
class HeapBuffer {
    static_assert(std::is_pod<POD_TYPE>::value, "Myst be POD-type");
public:
    HeapBuffer(unsigned long ItemsCnt) : m_buf{nullptr}, m_bytes_cnt{0},
        m_items_cnt{0}, m_hHeap(GetProcessHeap()) {
        const ::DWORD tmp = ItemsCnt * sizeof(POD_TYPE);
        if( m_hHeap )
            m_buf = static_cast<POD_TYPE*>(::HeapAlloc(m_hHeap, 0, tmp));
        if( m_buf ) {
            m_bytes_cnt = tmp;
            m_items_cnt = ItemsCnt;
        }
    }
    HeapBuffer(const HeapBuffer& ) = delete;
    HeapBuffer& operator=(const HeapBuffer& ) = delete;

    // isValid() ,   get()   
    // -- .
    bool isValid() const {
        return (m_buf != nullptr);
    }

    POD_TYPE* get() {
        return m_buf;
    }
    unsigned long ItemsCount() const {
        return m_items_cnt;
    }
    unsigned long BytesCount() const {
        return m_bytes_cnt;
    }
    bool FreeAndNull() const {
        //  (  )
        if( !m_buf )
            return true;
        //     
        const bool res{HeapFree(m_hHeap, 0, m_buf)};
        //   :
        m_buf = nullptr;
        m_bytes_cnt = 0;
        m_items_cnt = 0;
        return res; //     
        //    
    }
    ~HeapBuffer() {
        if( m_buf )
            ::HeapFree(m_hHeap, 0, m_buf);
    }
private:
    POD_TYPE *m_buf;
    ::DWORD m_bytes_cnt;
    ::DWORD m_items_cnt;
    const ::HANDLE m_hHeap;
};

class FileOutput {
public:
    FileOutput(std::string_view filename) : m_isValid{false},
        m_hFile{file_creator(filename)} {
        if( INVALID_HANDLE_VALUE != m_hFile )
            m_isValid = true;
    }
    FileOutput(const FileOutput& ) = delete;
    FileOutput& operator=(const FileOutput& ) = delete;
    ~FileOutput() {
        if( m_isValid )
            CloseHandle(m_hFile);
    }
    bool isValid() const {
        return m_isValid;
    }
    bool close() const; //   (  )
    bool write(const char* chars_buffer, unsigned long chars_to_write);
private:
    static HANDLE file_creator(std::string_view fName){
        return CreateFileA(fName.data(), GENERIC_WRITE, 0, NULL,
            CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    }
    mutable bool m_isValid;
    const HANDLE m_hFile;
};

#endif // HELPER_H
